from ._representations import (GenericWeakLabel, ContinuousWeakLabel, DiscreteWeakLabel, FuzzyLabel,
                              DiscreteFuzzyLabel, DiscreteSetLabel, IntervalFuzzyLabel, IntervalLabel)

__all__ = [
    "GenericWeakLabel",
    "ContinuousWeakLabel",
    "DiscreteWeakLabel",
    "FuzzyLabel",
    "DiscreteFuzzyLabel",
    "DiscreteSetLabel",
    "IntervalFuzzyLabel",
    "IntervalLabel"
]